"""Provides an ExperimentApplication class that manages the execution of an experiment."""

import time
from contextlib import contextmanager
from pathlib import Path
from typing import Optional, Union

from madsci.client.data_client import DataClient
from madsci.client.event_client import EventClient
from madsci.client.experiment_client import ExperimentClient
from madsci.client.resource_client import ResourceClient
from madsci.client.workcell_client import WorkcellClient
from madsci.common.exceptions import ExperimentCancelledError, ExperimentFailedError
from madsci.common.types.condition_types import Condition
from madsci.common.types.experiment_types import (
    Experiment,
    ExperimentDesign,
    ExperimentStatus,
)
from madsci.common.types.location_types import Location
from madsci.common.types.resource_types import Resource
from madsci.common.utils import threaded_daemon
from pydantic import AnyUrl
from rich import print


class ExperimentApplication:
    """
    An experiment application that helps manage the execution of an experiment.

    You can either use this class as a base class for your own application class, or create an instance of it to manage the execution of an experiment.
    """

    experiment: Optional[Experiment] = None
    """The current experiment being run."""
    experiment_design: Optional[ExperimentDesign] = None
    """The design of the experiment."""
    url: AnyUrl = AnyUrl("http://localhost:8002")
    """The URL of the experiment manager server."""
    logger = EventClient()
    """The event logger for the experiment."""

    def __init__(
        self,
        url: Optional[AnyUrl] = None,
        experiment_design: Optional[Union[str, Path, ExperimentDesign]] = None,
        experiment: Optional[Experiment] = None,
    ) -> "ExperimentApplication":
        """Initialize the experiment application. You can provide an experiment design to use for creating new experiments, or an existing experiment to continue."""
        self.url = AnyUrl(url) if url else self.url
        if experiment_design:
            self.experiment_design = experiment_design
        if isinstance(self.experiment_design, (str, Path)):
            self.experiment_design = ExperimentDesign.from_yaml(self.experiment_design)

        self.experiment = experiment if experiment else self.experiment

        self.experiment_client = ExperimentClient(url=self.url)

        self.workcell_client = WorkcellClient(
            self.experiment_client.workcell_client_url
        )
        self.resource_client = ResourceClient(
            self.experiment_client.resource_client_url
        )
        self.data_client = DataClient(self.experiment_client.data_client_url)
        if self.experiment_design and self.experiment_design.event_client_config:
            self.logger = EventClient(config=self.experiment_design.event_client_config)

    @classmethod
    def start_new(
        cls, url: AnyUrl, experiment_design: ExperimentDesign
    ) -> "ExperimentApplication":
        """Create a new experiment application with a new experiment."""
        self = cls(url=url, experiment_design=experiment_design)
        self.start_experiment_run()
        return self

    @classmethod
    def continue_experiment(
        cls, url: AnyUrl, experiment: Experiment
    ) -> "ExperimentApplication":
        """Create a new experiment application with an existing experiment."""
        self = cls(url=url, experiment=experiment)
        self.experiment_client.continue_experiment(
            experiment_id=experiment.experiment_id
        )
        return self

    def start_experiment_run(
        self, run_name: Optional[str] = None, run_description: Optional[str] = None
    ) -> None:
        """Sends the ExperimentDesign to the server to register a new experimental run."""

        self.experiment = self.experiment_client.start_experiment(
            experiment_design=self.experiment_design,
            run_name=run_name,
            run_description=run_description,
        )
        self.logger.log_info(
            f"Started run '{self.experiment.run_name}' ({self.experiment.experiment_id}) of experiment '{self.experiment.experiment_design.experiment_name}'"
        )
        passed_checks = False
        while not passed_checks:
            passed_checks = True
            for condition in self.experiment_design.resource_conditions:
                passed = self.evaluate_condition(condition)
                passed_checks = passed_checks and passed
                print(f"Check {condition.condition_name}: {passed}")
            if not passed_checks:
                val = input("Check failed, retry?")
                if val == "n":
                    break

        self.workcell_client.ownership_info.experiment_id = (
            self.experiment.experiment_id
        )
        self.data_client.ownership_info.experiment_id = self.experiment.experiment_id
        self.resource_client.ownership_info.experiment_id = (
            self.experiment.experiment_id
        )

    def end_experiment(self, status: Optional[ExperimentStatus] = None) -> None:
        """End the experiment."""
        self.experiment = self.experiment_client.end_experiment(
            experiment_id=self.experiment.experiment_id,
            status=status,
        )
        self.logger.log_info(
            f"Ended run '{self.experiment.run_name}' ({self.experiment.experiment_id}) of experiment '{self.experiment.experiment_design.experiment_name}'"
        )

    def pause_experiment(self) -> None:
        """Pause the experiment."""
        self.experiment = self.experiment_client.pause_experiment(
            experiment_id=self.experiment.experiment_id
        )
        self.logger.log_info(
            f"Paused run '{self.experiment.run_name}' ({self.experiment.experiment_id}) of experiment '{self.experiment.experiment_design.experiment_name}'"
        )

    def cancel_experiment(self) -> None:
        """Cancel the experiment."""
        self.experiment = self.experiment_client.cancel_experiment(
            experiment_id=self.experiment.experiment_id
        )
        self.logger.log_info(
            f"Cancelled run '{self.experiment.run_name}' ({self.experiment.experiment_id}) of experiment '{self.experiment.experiment_design.experiment_name}'"
        )

    def fail_experiment(self) -> None:
        """Fail the experiment."""
        self.experiment = self.experiment_client.end_experiment(
            experiment_id=self.experiment.experiment_id,
            status=ExperimentStatus.FAILED,
        )
        self.logger.log_info(
            f"Failed run '{self.experiment.run_name}' ({self.experiment.experiment_id}) of experiment '{self.experiment.experiment_design.experiment_name}'"
        )

    @contextmanager
    def manage_experiment(
        self, run_name: Optional[str] = None, run_description: Optional[str] = None
    ) -> contextmanager:
        """Context manager to start and end an experiment."""
        self.start_experiment_run(run_name=run_name, run_description=run_description)
        try:
            yield
        finally:
            self.end_experiment()

    @threaded_daemon
    def loop(self) -> None:
        """Function that runs the experimental loop. This should be overridden by subclasses."""
        raise NotImplementedError

    def check_experiment_status(self) -> None:
        """
        Update and check the status of the current experiment.

        Raises an exception if the experiment has been cancelled or failed.
        If the experiment has been paused, this function will wait until the experiment is resumed.

        Raises:
            ExperimentCancelledError: If the experiment has been cancelled.
            ExperimentFailedError: If the experiment has failed.
        """
        self.experiment = self.experiment_client.get_experiment(
            experiment_id=self.experiment.experiment_id
        )
        exception = None
        if self.experiment.status == ExperimentStatus.PAUSED:
            self.logger.log_warning(
                f"Experiment '{self.experiment.experiment_design.experiment_name}' has been paused."
            )
            while True:
                time.sleep(5)
                self.experiment = self.experiment_client.get_experiment(
                    experiment_id=self.experiment.experiment_id
                )
                if self.experiment.status != ExperimentStatus.PAUSED:
                    break
        if self.experiment.status == ExperimentStatus.CANCELLED:
            exception = ExperimentCancelledError(
                "Experiment manager reports that the experiment has been cancelled."
            )
        elif self.experiment.status == ExperimentStatus.FAILED:
            exception = ExperimentFailedError(
                "Experiment manager reports that the experiment has failed."
            )

        if exception:
            self.logger.log_error(exception.message)
            raise exception

    def get_resource_from_condition(self, condition: Condition) -> Resource | None:
        """gets a resource from a condition"""
        resource = None
        if condition.resource_id:
            resource = self.resource_client.get_resource(condition.resource_id)
        elif condition.resource_name:
            resource = self.resource_client.query_resource(
                resource_name=condition.resource_name, multiple=False
            )
        if resource is None:
            raise (Exception("Invalid Identifier for Resource"))
        return resource

    def check_resource_field(self, resource: Resource, condition: Condition) -> bool:
        """check if a resource meets a condition"""
        if condition.operator == "is_greater_than":
            return getattr(resource, condition.field) > condition.target_value
        if condition.operator == "is_less_than":
            return getattr(resource, condition.field) < condition.target_value
        if condition.operator == "is_equal_to":
            return getattr(resource, condition.field) == condition.target_value
        if condition.operator == "is_greater_than_or_equal_to":
            return getattr(resource, condition.field) >= condition.target_value
        if condition.operator == "is_less_than_or_equal_to":
            return getattr(resource, condition.field) < condition.target_value
        return False

    def get_location_from_condition(self, condition: Condition) -> Location:
        """get the location referenced by a condition"""
        location = None
        if condition.location_name:
            location = next(
                location
                for location in self.workcell_client.get_locations()
                if location.location_name == condition.location_name
            )
        elif condition.location_id:
            location = self.workcell_client.get_location(condition.location_id)
        if location is None:
            raise (Exception("Invalid Identifier for Location"))
        return location

    def resource_at_key(self, resource: Resource, condition: Condition) -> bool:
        """return if a resource is in a location at condition.key"""
        if isinstance(resource.children, list):
            if len(resource.children) > int(condition.key):
                if condition.resource_type:
                    return (
                        resource.children[condition.key].resource_type
                        == condition.resource_type
                    )
                return True
            return False
        if str(condition.key) in resource.children:
            if condition.resource_type:
                return (
                    resource.children[str(condition.key)].resource_type
                    == condition.resource_type
                )
            return True
        return False

    def evaluate_condition(self, condition: Condition) -> bool:
        """evaluate a condition"""
        if condition.condition_type == "resource_present":
            location = self.get_location_from_condition(condition)
            resource = self.resource_client.get_resource(location.resource_id)
            return self.resource_at_key(resource, condition)
        if condition.condition_type == "no_resource_present":
            location = self.get_location_from_condition(condition)
            resource = self.resource_client.get_resource(location.resource_id)
            return not self.resource_at_key(resource, condition)

        if condition.condition_type == "resource_field_check":
            resource = self.get_resource_from_condition(condition)
            return self.check_resource_field(resource, condition)

        if condition.condition_type == "resource_child_field_check":
            resource = self.get_resource_from_condition(condition)
            if isinstance(resource.children, list):
                if len(resource.children) > int(condition.key):
                    resource_child = resource.children[int(condition.key)]
                else:
                    raise (Exception("Invalid Key for Resource Child"))
            elif condition.key not in resource.children:
                raise (Exception("Invalid Key for Resource Child"))
            else:
                resource_child = resource.children[condition.key]
            return self.check_resource_field(resource_child, condition)
        return False


if __name__ == "__main__":
    import datetime

    class MyExperimentApplication(ExperimentApplication):
        """An example experiment application."""

        experiment_design = ExperimentDesign(
            experiment_name="My Example Experiment",
            experiment_description="An example experimental design",
        )
        url = "http://localhost:8002"

        def loop(self, iterations: int = 10) -> None:
            """Run the experiment loop."""
            for i in range(iterations):
                time.sleep(10)
                self.check_experiment_status()
                self.logger.log_info(f"Running experiment loop {i}")

    experiment_app = MyExperimentApplication()
    current_time = datetime.datetime.now()
    with experiment_app.manage_experiment(
        run_name=f"My Experiment Run {current_time}",
        run_description=f"Run for my example experiment, started at ~{current_time}",
    ):
        experiment_app.loop(iterations=10)



"""Command Line Interface for the MADSci client."""

import click
from madsci.client.cli.lab_cli import lab
from madsci.client.cli.manager_cli import manager
from madsci.client.cli.node_cli import node
from madsci.client.cli.resources_cli import resource
from madsci.client.cli.workcell_cli import workcell
from rich.console import Console
from trogon import tui

console = Console()


@tui()
@click.group()
@click.option(
    "--quiet",
    "-q",
    is_flag=True,
    help="Run in quiet mode, skipping prompts.",
)
def root_cli(quiet: bool = False) -> None:
    """MADSci command line interface."""


@root_cli.command()
def version() -> None:
    """Display the MADSci client version."""
    console.print("MADSci Client v0.1.0")


root_cli.add_command(lab)
root_cli.add_command(workcell)
root_cli.add_command(node)
root_cli.add_command(resource)
root_cli.add_command(manager)

if __name__ == "__main__":
    tui(root_cli, auto_envvar_prefix="MADSCI_CLI_")



"""Unit tests for WorkcellClient."""

from collections.abc import Generator
from typing import Any
from unittest.mock import patch

import pytest
from fastapi.testclient import TestClient
from madsci.client.workcell_client import WorkcellClient
from madsci.common.types.base_types import new_ulid_str
from madsci.common.types.location_types import Location, LocationDefinition
from madsci.common.types.workcell_types import WorkcellDefinition, WorkcellState
from madsci.workcell_manager.workcell_server import (
    WorkflowDefinition,
    create_workcell_server,
)
from pymongo.synchronous.database import Database
from pytest_mock_resources import (
    MongoConfig,
    RedisConfig,
    create_mongo_fixture,
    create_redis_fixture,
)
from redis import Redis
from requests import Response


# Create a Redis server fixture for testing
@pytest.fixture(scope="session")
def pmr_redis_config() -> RedisConfig:
    """Configure the Redis server."""
    return RedisConfig(image="redis:7.4")


@pytest.fixture(scope="session")
def pmr_mongo_config() -> MongoConfig:
    """Congifure the MongoDB fixture."""
    return MongoConfig(image="mongo:8.0")


redis_server = create_redis_fixture()
mongo_server = create_mongo_fixture()


@pytest.fixture
def workcell() -> WorkcellDefinition:
    """Fixture for creating a WorkcellDefinition."""
    return WorkcellDefinition(
        workcell_name="Test Workcell",
        locations=[
            LocationDefinition(location_name="test_location"),
        ],
    )


@pytest.fixture
def test_client(
    workcell: WorkcellDefinition, redis_server: Redis, mongo_server: Database
) -> Generator[TestClient, None, None]:
    """Workcell Server Test Client Fixture."""
    app = create_workcell_server(
        workcell, redis_server, mongo_server, start_engine=False
    )
    client = TestClient(app)
    with client:
        yield client


@pytest.fixture
def client(test_client: TestClient) -> Generator[WorkcellClient, None, None]:
    """Fixture for WorkcellClient patched to use TestClient."""
    with patch("madsci.client.workcell_client.requests") as mock_requests:

        def post_no_timeout(*args: Any, **kwargs: Any) -> Response:
            kwargs.pop("timeout", None)
            return test_client.post(*args, **kwargs)

        mock_requests.post.side_effect = post_no_timeout

        def get_no_timeout(*args: Any, **kwargs: Any) -> Response:
            kwargs.pop("timeout", None)
            return test_client.get(*args, **kwargs)

        mock_requests.get.side_effect = get_no_timeout

        def delete_no_timeout(*args: Any, **kwargs: Any) -> Response:
            kwargs.pop("timeout", None)
            return test_client.delete(*args, **kwargs)

        mock_requests.delete.side_effect = delete_no_timeout

        yield WorkcellClient(workcell_manager_url="http://testserver")


def test_get_nodes(client: WorkcellClient) -> None:
    """Test retrieving nodes from the workcell."""
    response = client.add_node("node1", "http://node1/")
    assert response["node_url"] == "http://node1/"
    nodes = client.get_nodes()
    assert "node1" in nodes
    assert nodes["node1"]["node_url"] == "http://node1/"


def test_get_node(client: WorkcellClient) -> None:
    """Test retrieving a specific node."""
    client.add_node("node1", "http://node1/")
    node = client.get_node("node1")
    assert node["node_url"] == "http://node1/"


def test_add_node(client: WorkcellClient) -> None:
    """Test adding a node to the workcell."""
    node = client.add_node("node1", "http://node1/")
    assert node["node_url"] == "http://node1/"


def test_get_active_workflows(client: WorkcellClient) -> None:
    """Test retrieving workflows."""
    workflows = client.get_active_workflows()
    assert isinstance(workflows, dict)


def test_get_archived_workflows(client: WorkcellClient) -> None:
    """Test retrieving workflows."""
    workflows = client.get_archived_workflows(30)
    assert isinstance(workflows, dict)


def test_get_workflow_queue(client: WorkcellClient) -> None:
    """Test retrieving the workflow queue."""
    queue = client.get_workflow_queue()
    assert isinstance(queue, list)


def test_get_workcell_state(client: WorkcellClient) -> None:
    """Test retrieving the workcell state."""
    state = client.get_workcell_state()
    assert isinstance(state, WorkcellState)


def test_pause_workflow(client: WorkcellClient) -> None:
    """Test pausing a workflow."""
    workflow = client.submit_workflow(
        WorkflowDefinition(name="Test Workflow"), None, await_completion=False
    )
    paused_workflow = client.pause_workflow(workflow.workflow_id)
    assert paused_workflow.status.paused is True


def test_resume_workflow(client: WorkcellClient) -> None:
    """Test resuming a workflow."""
    workflow = client.submit_workflow(
        WorkflowDefinition(name="Test Workflow"), {}, await_completion=False
    )
    client.pause_workflow(workflow.workflow_id)
    resumed_workflow = client.resume_workflow(workflow.workflow_id)
    assert resumed_workflow.status.paused is False


def test_cancel_workflow(client: WorkcellClient) -> None:
    """Test canceling a workflow."""
    workflow = client.submit_workflow(
        WorkflowDefinition(name="Test Workflow"), {}, await_completion=False
    )
    canceled_workflow = client.cancel_workflow(workflow.workflow_id)
    assert canceled_workflow.status.cancelled is True


def test_get_locations(client: WorkcellClient) -> None:
    """Test retrieving locations."""
    locations = client.get_locations()
    assert isinstance(locations, list)
    assert len(locations) == 1
    assert locations[0].location_name == "test_location"


def test_get_location(client: WorkcellClient) -> None:
    """Test retrieving a specific location."""
    location_id = client.get_locations()[0].location_id
    fetched_location = client.get_location(location_id)
    assert fetched_location.location_id == location_id


def test_add_location(client: WorkcellClient) -> None:
    """Test adding a location."""
    location = Location(location_name="test_location2")
    added_location = client.add_location(location)
    assert added_location.location_id == location.location_id
    assert added_location.location_name == location.location_name


def test_attach_resource_to_location(client: WorkcellClient) -> None:
    """Test attaching a resource to a location."""
    location = Location(location_name="test_location3")
    client.add_location(location)
    mock_resource_id = new_ulid_str()
    updated_location = client.attach_resource_to_location(
        location.location_id, mock_resource_id
    )
    assert updated_location.resource_id == mock_resource_id



"""Client for performing workcell actions"""

import json
import time
from pathlib import Path, PosixPath, PurePath, WindowsPath
from typing import Any, Optional, Union

import requests
from madsci.common.data_manipulation import value_substitution, walk_and_replace
from madsci.common.exceptions import WorkflowFailedError
from madsci.common.types.auth_types import OwnershipInfo
from madsci.common.types.base_types import new_ulid_str
from madsci.common.types.location_types import Location
from madsci.common.types.node_types import Node
from madsci.common.types.workcell_types import WorkcellState
from madsci.common.types.workflow_types import (
    Workflow,
    WorkflowDefinition,
)
from madsci.common.utils import PathLike
from rich import print


class WorkcellClient:
    """A client for interacting with the Workcell Manager to perform various actions."""

    def __init__(
        self,
        workcell_manager_url: str,
        working_directory: str = "./",
        ownership_info: Optional[OwnershipInfo] = None,
    ) -> None:
        """
        Initialize the WorkcellClient.

        Parameters
        ----------
        workcell_manager_url : str
            The base URL of the Workcell Manager.
        working_directory : str, optional
            The directory to look for relative paths. Defaults to "./".
        ownership_info : Optional[OwnershipInfo], optional
            Ownership information for workflows, by default None.
        """
        self.url = workcell_manager_url
        self.working_directory = Path(working_directory).expanduser()
        self.ownership_info = ownership_info
        if self.ownership_info is None:
            self.ownership_info = OwnershipInfo()
        if str(self.url).endswith("/"):
            self.url = str(self.url)[:-1]

    def query_workflow(self, workflow_id: str) -> Optional[Workflow]:
        """
        Check the status of a workflow using its ID.

        Parameters
        ----------
        workflow_id : str
            The ID of the workflow to query.

        Returns
        -------
        Optional[Workflow]
            The workflow object if found, otherwise None.
        """
        url = f"{self.url}/workflow/{workflow_id}"
        response = requests.get(url, timeout=10)

        response.raise_for_status()
        return Workflow(**response.json())

    def submit_workflow(
        self,
        workflow: Union[PathLike, WorkflowDefinition],
        parameters: Optional[dict[str, Any]] = None,
        validate_only: bool = False,
        await_completion: bool = True,
        raise_on_failed: bool = True,
        raise_on_cancelled: bool = True,
    ) -> Workflow:
        """
        Submit a workflow to the Workcell Manager.

        Parameters
        ----------
        workflow: Union[PathLike, WorkflowDefinition],
            Either a WorkflowDefinition or a path to a YAML file of one.
        parameters: Optional[dict[str, Any]] = None,
            Parameters to be inserted into the workflow.
        validate_only : bool, optional
            If True, only validate the workflow without submitting, by default False.
        await_completion : bool, optional
            If True, wait for the workflow to complete, by default True.
        raise_on_failed : bool, optional
            If True, raise an exception if the workflow fails, by default True.
        raise_on_cancelled : bool, optional
            If True, raise an exception if the workflow is cancelled, by default True.

        Returns
        -------
        Workflow
            The submitted workflow object.
        """
        if isinstance(workflow, (Path, str)):
            workflow = WorkflowDefinition.from_yaml(workflow)
        else:
            workflow = WorkflowDefinition.model_validate(workflow)
        insert_parameter_values(
            workflow=workflow, parameters=parameters if parameters else {}
        )
        files = self._extract_files_from_workflow(workflow)
        url = self.url + "/workflow"
        response = requests.post(
            url,
            data={
                "workflow": workflow.model_dump_json(),
                "parameters": json.dumps(parameters) if parameters else None,
                "validate_only": validate_only,
                "ownership_info": self.ownership_info.model_dump_json()
                if self.ownership_info
                else None,
            },
            files={
                (
                    "files",
                    (
                        str(Path(path).name),
                        Path.open(Path(path).expanduser(), "rb"),
                    ),
                )
                for _, path in files.items()
            },
            timeout=10,
        )
        response.raise_for_status()
        if not await_completion:
            return Workflow(**response.json())
        return self.await_workflow(
            response.json()["workflow_id"],
            raise_on_cancelled=raise_on_cancelled,
            raise_on_failed=raise_on_failed,
        )

    start_workflow = submit_workflow

    def _extract_files_from_workflow(
        self, workflow: WorkflowDefinition
    ) -> dict[str, Path]:
        """
        Extract file paths from a workflow definition.

        Parameters
        ----------
        workflow : WorkflowDefinition
            The workflow definition object.

        Returns
        -------
        dict[str, Path]
            A dictionary mapping unique file names to their paths.
        """
        files = {}
        for step in workflow.steps:
            if step.files:
                for file, path in step.files.items():
                    unique_filename = f"{new_ulid_str()}_{file}"
                    files[unique_filename] = path
                    if not Path(files[unique_filename]).is_absolute():
                        files[unique_filename] = (
                            self.working_directory / files[unique_filename]
                        )
                    step.files[file] = Path(files[unique_filename]).name
        return files

    def submit_workflow_sequence(
        self, workflows: list[str], parameters: list[dict[str, Any]]
    ) -> list[Workflow]:
        """
        Submit a sequence of workflows to run in order.

        Parameters
        ----------
        workflows : list[str]
            A list of workflow definitions in YAML format.
        parameters : list[dict[str, Any]]
            A list of parameter dictionaries for each workflow.

        Returns
        -------
        list[Workflow]
            A list of submitted workflow objects.
        """
        wfs = []
        for i in range(len(workflows)):
            wf = self.submit_workflow(
                workflows[i], parameters[i], await_completion=True
            )
            wfs.append(wf)
        return wfs

    def submit_workflow_batch(
        self, workflows: list[str], parameters: list[dict[str, Any]]
    ) -> list[Workflow]:
        """
        Submit a batch of workflows to run concurrently.

        Parameters
        ----------
        workflows : list[str]
            A list of workflow definitions in YAML format.
        parameters : list[dict[str, Any]]
            A list of parameter dictionaries for each workflow.

        Returns
        -------
        list[Workflow]
            A list of completed workflow objects.
        """
        id_list = []
        for i in range(len(workflows)):
            response = self.submit_workflow(
                workflows[i], parameters[i], await_completion=False
            )
            id_list.append(response.json()["workflow_id"])
        finished = False
        while not finished:
            flag = True
            wfs = []
            for id in id_list:
                wf = self.query_workflow(id)
                flag = flag and (wf.status.terminal)
                wfs.append(wf)
            finished = flag
        return wfs

    def retry_workflow(self, workflow_id: str, index: int = -1) -> dict:
        """
        Retry a workflow from a specific step.

        Parameters
        ----------
        workflow_id : str
            The ID of the workflow to retry.
        index : int, optional
            The step index to retry from, by default -1 (retry the entire workflow).

        Returns
        -------
        dict
            The response from the Workcell Manager.
        """
        url = f"{self.url}/workflow/{workflow_id}/retry"
        response = requests.post(
            url,
            params={
                "workflow_id": workflow_id,
                "index": index,
            },
            timeout=10,
        )
        return response.json()

    def resubmit_workflow(
        self,
        workflow_id: str,
        await_completion: bool = True,
        raise_on_failed: bool = True,
        raise_on_cancelled: bool = True,
    ) -> Workflow:
        """
        Resubmit a workflow as a new workflow with a new ID.

        Parameters
        ----------
        workflow_id : str
            The ID of the workflow to resubmit.
        await_completion : bool, optional
            If True, wait for the workflow to complete, by default True.
        raise_on_failed : bool, optional
            If True, raise an exception if the workflow fails, by default True.
        raise_on_cancelled : bool, optional
            If True, raise an exception if the workflow is cancelled, by default True.

        Returns
        -------
        Workflow
            The resubmitted workflow object.
        """
        url = f"{self.url}/workflow/{workflow_id}/resubmit"
        response = requests.get(url, timeout=10)
        new_wf = Workflow(**response.json())
        if await_completion:
            return self.await_workflow(
                new_wf.workflow_id,
                raise_on_failed=raise_on_failed,
                raise_on_cancelled=raise_on_cancelled,
            )
        return new_wf

    def await_workflow(
        self,
        workflow_id: str,
        raise_on_failed: bool = True,
        raise_on_cancelled: bool = True,
    ) -> Workflow:
        """
        Wait for a workflow to complete.

        Parameters
        ----------
        workflow_id : str
            The ID of the workflow to wait for.
        raise_on_failed : bool, optional
            If True, raise an exception if the workflow fails, by default True.
        raise_on_cancelled : bool, optional
            If True, raise an exception if the workflow is cancelled, by default True.

        Returns
        -------
        Workflow
            The completed workflow object.
        """
        prior_status = None
        prior_index = None
        while True:
            wf = self.query_workflow(workflow_id)
            status = wf.status
            step_index = wf.status.current_step_index
            if prior_status != status or prior_index != step_index:
                if step_index < len(wf.steps):
                    step_name = wf.steps[step_index].name
                else:
                    step_name = "Workflow End"
                # TODO: Improve progress reporting
                print(
                    f"\n{wf.name}['{step_name}']: {wf.status.description}",
                    end="",
                    flush=True,
                )
            else:
                print(".", end="", flush=True)
            time.sleep(1)
            if wf.status.terminal:
                print()
                break
            prior_status = status
            prior_index = step_index
        if wf.status.failed and raise_on_failed:
            raise WorkflowFailedError(
                f"Workflow {wf.name} ({wf.workflow_id}) failed on step {wf.status.current_step_index}: '{wf.steps[wf.status.current_step_index].name}'."
            )
        if wf.status.cancelled and raise_on_cancelled:
            raise WorkflowFailedError(
                f"Workflow {wf.name} ({wf.workflow_id}) was cancelled on step {wf.status.current_step_index}: '{wf.steps[wf.status.current_step_index].name}'."
            )
        return wf

    def get_nodes(self) -> dict[str, Node]:
        """
        Get all nodes in the workcell.

        Returns
        -------
        dict[str, Node]
            A dictionary of node names and their details.
        """
        url = f"{self.url}/nodes"
        response = requests.get(url, timeout=10)
        return response.json()

    def get_node(self, node_name: str) -> Node:
        """
        Get details of a specific node.

        Parameters
        ----------
        node_name : str
            The name of the node.

        Returns
        -------
        Node
            The node details.
        """
        url = f"{self.url}/node/{node_name}"
        response = requests.get(url, timeout=10)
        return response.json()

    def add_node(
        self,
        node_name: str,
        node_url: str,
        node_description: str = "A Node",
        permanent: bool = False,
    ) -> Node:
        """
        Add a node to the workcell.

        Parameters
        ----------
        node_name : str
            The name of the node.
        node_url : str
            The URL of the node.
        node_description : str, optional
            A description of the node, by default "A Node".
        permanent : bool, optional
            If True, add the node permanently, by default False.

        Returns
        -------
        Node
            The added node details.
        """
        url = f"{self.url}/node"
        response = requests.post(
            url,
            params={
                "node_name": node_name,
                "node_url": node_url,
                "node_description": node_description,
                "permanent": permanent,
            },
            timeout=10,
        )
        return response.json()

    def get_active_workflows(self) -> dict[str, Workflow]:
        """
        Get all workflows from the Workcell Manager.

        Returns
        -------
        dict[str, Workflow]
            A dictionary of workflow IDs and their details.
        """
        url = f"{self.url}/workflows/active"
        response = requests.get(url, timeout=100)
        response.raise_for_status()
        workflow_dict = response.json()
        if not isinstance(workflow_dict, dict):
            raise ValueError(
                f"Expected a dictionary of workflows, but got {type(workflow_dict)}."
            )
        return {
            key: Workflow.model_validate(value) for key, value in workflow_dict.items()
        }

    def get_archived_workflows(self, number: int = 20) -> dict[str, Workflow]:
        """
        Get all workflows from the Workcell Manager.

        Returns
        -------
        dict[str, Workflow]
            A dictionary of workflow IDs and their details.
        """
        url = f"{self.url}/workflows/archived"
        response = requests.get(url, params={"number": number}, timeout=100)
        response.raise_for_status()
        workflow_dict = response.json()
        if not isinstance(workflow_dict, dict):
            raise ValueError(
                f"Expected a dictionary of workflows, but got {type(workflow_dict)}."
            )
        return {
            key: Workflow.model_validate(value) for key, value in workflow_dict.items()
        }

    def get_workflow_queue(self) -> list[Workflow]:
        """
        Get the workflow queue from the workcell.

        Returns
        -------
        list[Workflow]
            A list of queued workflows.
        """
        url = f"{self.url}/workflows/queue"
        response = requests.get(url, timeout=10)
        response.raise_for_status()
        return [Workflow.model_validate(wf) for wf in response.json()]

    def get_workcell_state(self) -> WorkcellState:
        """
        Get the full state of the workcell.

        Returns
        -------
        WorkcellState
            The current state of the workcell.
        """
        url = f"{self.url}/state"
        response = requests.get(url, timeout=10)
        response.raise_for_status()
        return WorkcellState.model_validate(response.json())

    def pause_workflow(self, workflow_id: str) -> Workflow:
        """
        Pause a workflow.

        Parameters
        ----------
        workflow_id : str
            The ID of the workflow to pause.

        Returns
        -------
        Workflow
            The paused workflow object.
        """
        url = f"{self.url}/workflow/{workflow_id}/pause"
        response = requests.post(url, timeout=10)
        response.raise_for_status()
        return Workflow.model_validate(response.json())

    def resume_workflow(self, workflow_id: str) -> Workflow:
        """
        Resume a paused workflow.

        Parameters
        ----------
        workflow_id : str
            The ID of the workflow to resume.

        Returns
        -------
        Workflow
            The resumed workflow object.
        """
        url = f"{self.url}/workflow/{workflow_id}/resume"
        response = requests.post(url, timeout=10)
        response.raise_for_status()
        return Workflow.model_validate(response.json())

    def cancel_workflow(self, workflow_id: str) -> Workflow:
        """
        Cancel a workflow.

        Parameters
        ----------
        workflow_id : str
            The ID of the workflow to cancel.

        Returns
        -------
        Workflow
            The cancelled workflow object.
        """
        url = f"{self.url}/workflow/{workflow_id}/cancel"
        response = requests.post(url, timeout=10)
        response.raise_for_status()
        return Workflow.model_validate(response.json())

    def get_locations(self) -> list[Location]:
        """
        Get all locations in the workcell.

        Returns
        -------
        list[Location]
            A list of locations.
        """
        url = f"{self.url}/locations"
        response = requests.get(url, timeout=10)
        response.raise_for_status()
        return [Location.model_validate(loc) for loc in response.json().values()]

    def get_location(self, location_id: str) -> Location:
        """
        Get details of a specific location.

        Parameters
        ----------
        location_id : str
            The ID of the location.

        Returns
        -------
        Location
            The location details.
        """
        url = f"{self.url}/location/{location_id}"
        response = requests.get(url, timeout=10)
        response.raise_for_status()
        return Location.model_validate(response.json())

    def add_location(self, location: Location) -> Location:
        """
        Add a location to the workcell.

        Parameters
        ----------
        location : Location
            The location object to add.

        Returns
        -------
        Location
            The added location details.
        """
        url = f"{self.url}/location"
        response = requests.post(
            url,
            json=location.model_dump(mode="json"),
            timeout=10,
        )
        response.raise_for_status()
        return Location.model_validate(response.json())

    def attach_resource_to_location(
        self, location_id: str, resource_id: str
    ) -> Location:
        """
        Attach a resource container to a location.

        Parameters
        ----------
        location_id : str
            The ID of the location.
        resource_id : str
            The ID of the resource to attach.

        Returns
        -------
        Location
            The updated location details.
        """
        url = f"{self.url}/location/{location_id}/attach_resource"
        response = requests.post(
            url,
            params={
                "resource_id": resource_id,
            },
            timeout=10,
        )
        response.raise_for_status()
        return Location.model_validate(response.json())

    def delete_location(self, location_id: str) -> None:
        """
        Delete a location from the workcell.

        Parameters
        ----------
        location_id : str
            The ID of the location to delete.

        Returns
        -------
        dict
            A dictionary indicating the deletion status.
        """
        url = f"{self.url}/location/{location_id}"
        response = requests.delete(url, timeout=10)
        response.raise_for_status()


def insert_parameter_values(
    workflow: WorkflowDefinition, parameters: dict[str, Any]
) -> Workflow:
    """Replace the parameter strings in the workflow with the provided values"""
    for param in workflow.parameters:
        if param.name not in parameters:
            if param.default:
                parameters[param.name] = param.default
            else:
                raise ValueError(
                    "Workflow parameter: "
                    + param.name
                    + " not provided, and no default value is defined."
                )
    for key, value in parameters.items():
        if isinstance(value, (Path, PurePath, WindowsPath, PosixPath)):
            parameters[key] = str(value)
    steps = []
    for step in workflow.steps:
        for key, val in iter(step):
            if type(val) is str:
                setattr(step, key, value_substitution(val, parameters))

        step.args = walk_and_replace(step.args, parameters)
        step.files = walk_and_replace(step.files, parameters)
        step.locations = walk_and_replace(step.locations, parameters)
        steps.append(step)
    workflow.steps = steps



"""Command Line Interface for managing MADSci Squid workcells."""

from pathlib import Path
from typing import Optional

import click
from click import Context
from madsci.client.cli.lab_cli import LabContext, find_lab
from madsci.common.types.workcell_types import WorkcellDefinition, WorkcellLink
from madsci.common.utils import (
    prompt_for_input,
    prompt_yes_no,
    save_model,
    search_for_file_pattern,
    to_snake_case,
)
from rich.console import Console
from rich.pretty import pprint

console = Console()


class WorkcellContext:
    """Context object for workcell commands."""

    def __init__(self) -> None:
        """Initialize the context object."""
        self.workcell: Optional[WorkcellDefinition] = None
        self.path: Optional[Path] = None
        self.lab: Optional[LabContext] = None
        self.quiet: bool = False


pass_workcell = click.make_pass_decorator(WorkcellContext)


def find_workcell(
    name: Optional[str],
    path: Optional[str],
    lab_context: Optional[LabContext] = None,
) -> WorkcellContext:
    """Find a workcell by name or path."""
    workcell_context = WorkcellContext()
    workcell_context.lab = lab_context

    if path:
        workcell_context.path = Path(path)
        if workcell_context.path.exists():
            workcell_context.workcell = WorkcellDefinition.from_yaml(path)
            return workcell_context

    # If we have a lab context, search in the lab directory first
    if lab_context and lab_context.path:
        workcell_files = search_for_file_pattern(
            "*.workcell.yaml",
            start_dir=lab_context.path.parent,
        )
        for workcell_file in workcell_files:
            workcell = WorkcellDefinition.from_yaml(workcell_file)
            if not name or workcell.name == name:
                workcell_context.path = Path(workcell_file)
                workcell_context.workcell = workcell
                return workcell_context

    # If not found in lab directory or no lab context, search everywhere
    workcell_files = search_for_file_pattern("*.workcell.yaml")
    for workcell_file in workcell_files:
        workcell = WorkcellDefinition.from_yaml(workcell_file)
        if not name or workcell.name == name:
            workcell_context.path = Path(workcell_file)
            workcell_context.workcell = workcell
            return workcell_context

    return workcell_context


@click.group()
@click.option("--name", "-n", type=str, help="Name of the workcell.")
@click.option("--path", "-p", type=str, help="Path to the workcell definition file.")
@click.option("--lab", "-l", type=str, help="Name or path of the lab to operate in.")
@click.pass_context
def workcell(
    ctx: Context,
    name: Optional[str],
    path: Optional[str],
    lab: Optional[str],
) -> None:
    """Manage workcells. Specify workcell by name or path."""
    lab_context = find_lab(name=lab, path=lab)
    ctx.obj = find_workcell(name=name, path=path, lab_context=lab_context)
    ctx.obj.quiet = ctx.parent.params.get("quiet")


@workcell.command()
@click.option("--name", "-n", type=str, help="The name of the workcell.")
@click.option(
    "--path",
    "-p",
    type=str,
    help="The path to the workcell definition file.",
)
@click.option("--description", "-d", type=str, help="The description of the workcell.")
@click.option("--yes", "-y", is_flag=True, help="Skip confirmation prompt.")
@click.pass_context
def create(
    ctx: Context,
    name: Optional[str],
    path: Optional[str],
    description: Optional[str],
    yes: bool,
) -> None:
    """Create a new workcell."""
    name = name if name else ctx.parent.params.get("name")
    name = (
        name
        if name
        else prompt_for_input("Workcell Name", required=True, quiet=ctx.obj.quiet)
    )
    description = (
        description
        if description
        else prompt_for_input("Workcell Description", quiet=ctx.obj.quiet)
    )

    workcell = WorkcellDefinition(name=name, description=description)
    console.print(workcell)

    path = path if path else ctx.parent.params.get("path")
    if not path:
        if ctx.obj.lab and ctx.obj.lab.path:
            # If we have a lab context, create in the lab directory
            path = (
                ctx.obj.lab.path.parent
                / "workcells"
                / f"{to_snake_case(name)}.workcell.yaml"
            )
        else:
            current_path = Path.cwd()
            if current_path.name == "workcells":
                path = current_path / f"{to_snake_case(name)}.workcell.yaml"
            else:
                path = (
                    current_path / "workcells" / f"{to_snake_case(name)}.workcell.yaml"
                )

        new_path = prompt_for_input(
            "Path to save Workcell Definition file",
            default=path,
            quiet=ctx.obj.quiet,
        )
        if new_path:
            path = Path(new_path)
    else:
        path = Path(path)

    if not path.expanduser().exists():
        path.expanduser().parent.mkdir(parents=True, exist_ok=True)
    save_model(path, workcell, overwrite_check=not ctx.obj.quiet and not yes)

    if (
        ctx.obj.lab
        and ctx.obj.lab.lab_def
        and (
            yes
            or ctx.obj.quiet
            or prompt_yes_no(
                f"Add workcell to lab [bold]{ctx.obj.lab.lab_def.name}[/] ([italic]{ctx.obj.lab.path}[/])?",
                default="yes",
            )
        )
    ):
        relative_path = path.relative_to(ctx.obj.lab.path.parent)
        if name not in ctx.obj.lab.lab_def.managers:
            ctx.obj.lab.lab_def.managers[name] = WorkcellLink(path=relative_path)
            save_model(ctx.obj.lab.path, ctx.obj.lab.lab_def, overwrite_check=False)


@workcell.command()
@pass_workcell
def list(ctx: WorkcellContext) -> None:
    """List all workcells."""
    search_dir = ctx.lab.path.parent if ctx.lab and ctx.lab.path else None
    workcell_files = search_for_file_pattern("*.workcell.yaml", start_dir=search_dir)

    if workcell_files:
        for workcell_file in sorted(set(workcell_files)):
            workcell = WorkcellDefinition.from_yaml(workcell_file)
            console.print(
                f"[bold]{workcell.name}[/]: {workcell.description} ({workcell_file})",
            )
    else:
        lab_context = " in lab directory" if ctx.lab and ctx.lab.path else ""
        print(  # noqa: T201
            f"No workcell definitions found{lab_context}, you can create one with 'madsci workcell create'",
        )


@workcell.command()
@pass_workcell
def info(ctx: WorkcellContext) -> None:
    """Get information about a workcell."""
    if ctx.workcell:
        pprint(ctx.workcell)
    else:
        print(  # noqa: T201
            "No workcell specified/found, please specify a workcell with --name or --path, or create a new workcell with 'madsci workcell create'",
        )


@workcell.command()
@click.option("--yes", "-y", is_flag=True, help="Skip confirmation prompt.")
@click.option("--name", "-n", type=str, help="The name of the workcell.")
@click.option(
    "--path",
    "-p",
    type=str,
    help="The path to the workcell definition file.",
)
@pass_workcell
def delete(
    ctx: WorkcellContext,
    yes: bool,
    name: Optional[str],
    path: Optional[str],
) -> None:
    """Delete a workcell."""
    if name or path:
        ctx.workcell = find_workcell(name=name, path=None, lab_context=ctx.lab).workcell
    if ctx.workcell and ctx.path:
        console.print(f"Deleting workcell: {ctx.workcell.name} ({ctx.path})")
        if yes or ctx.quiet or prompt_yes_no("Are you sure?", default="no"):
            ctx.path.unlink()
            console.print(f"Deleted {ctx.path}")
            if (
                (ctx.lab and ctx.lab.lab_def and yes)
                or ctx.quiet
                or prompt_yes_no(
                    f"Remove from lab [bold]{ctx.lab.lab_def.name}[/] ([italic]{ctx.lab.path}[/])?",
                    default="yes",
                )
            ) and ctx.workcell.name in ctx.lab.lab_def.managers:
                del ctx.lab.lab_def.managers[ctx.workcell.name]
                save_model(ctx.lab.path, ctx.lab.lab_def, overwrite_check=False)
    else:
        print(  # noqa: T201
            "No workcell specified/found, please specify a workcell with --name or --path, or create a new workcell with 'madsci workcell create'",
        )


@workcell.command()
@pass_workcell
def validate(ctx: WorkcellContext) -> None:
    """Validate a workcell definition file."""
    if ctx.workcell:
        console.print(ctx.workcell)
        return
    console.print(
        "No workcell specified, please specify a workcell with --name or --path, or create a new workcell with 'madsci workcell create'",
    )
    return



"""MADSci Event Handling."""

import contextlib
import inspect
import logging
from collections import OrderedDict
from pathlib import Path
from typing import Any, Optional, Union

import requests
from madsci.common.types.event_types import (
    Event,
    EventClientConfig,
    EventType,
)
from madsci.common.utils import threaded_task
from pydantic import ValidationError
from rich import print


class EventClient:
    """A logger and event handler for MADSci system components."""

    config: Optional[EventClientConfig] = None

    def __init__(
        self,
        config: Optional[EventClientConfig] = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the event logger. If no config is provided, use the default config.

        Keyword Arguments are used to override the values of the passed in/default config.
        """
        if kwargs:
            if config:
                [self.config.__setattr__(key, value) for key, value in kwargs.items()]
            else:
                self.config = EventClientConfig(**kwargs)
        if config is not None:
            self.config = config
        else:
            self.config = EventClientConfig()
        if self.config.name:
            self.name = self.config.name
        else:
            # * See if there's a calling class we can name after
            stack = inspect.stack()
            parent = stack[1][0]
            if calling_self := parent.f_locals.get("self"):
                self.name = calling_self.__class__.__name__
            else:
                # * No luck, name after EventClient
                self.name = __name__
        self.name = str(self.name)
        self.logger = logging.getLogger()
        self.log_dir = Path(self.config.log_dir).expanduser()
        self.log_dir.mkdir(parents=True, exist_ok=True)
        self.logfile = self.log_dir / f"{self.name}.log"
        self.logger.setLevel(self.config.log_level)
        for handler in self.logger.handlers:
            if isinstance(handler, logging.FileHandler) and handler.baseFilename == str(
                self.logfile
            ):
                self.logger.removeHandler(handler)
        file_handler = logging.FileHandler(filename=str(self.logfile), mode="a+")
        self.logger.addHandler(file_handler)
        self.event_server = self.config.event_server_url
        self.source = self.config.source
        self.log_debug(
            Event(
                event_type=EventType.LOG_INFO,
                event_data=f"Event logger {self.name} initialized.",
            )
        )

    def get_log(self) -> dict[str, Event]:
        """Read the log"""
        events = {}
        with self.logfile.open() as log:
            for line in log.readlines():
                try:
                    event = Event.model_validate_json(line)
                except ValidationError:
                    event = Event(event_type=EventType.UNKNOWN, event_data=line)
                events[event.event_id] = event
        return events

    def get_events(self, number: int = 100, level: int = -1) -> dict[str, Event]:
        """Query the event server for a certain number of recent events. If no event server is configured, query the log file instead."""
        if level == -1:
            level = self.logger.getEffectiveLevel()
        events = OrderedDict()
        if self.event_server:
            response = requests.get(
                self.event_server + "/events",
                timeout=10,
                params={"number": number, "level": level},
            )
            if not response.ok:
                response.raise_for_status()
            print(response.json())
            for key, value in response.json().items():
                events[key] = Event.model_validate(value)
            return dict(events)
        events = self.get_log()
        selected_events = {}
        for event in reversed(list(events.values())):
            selected_events[event.event_id] = event
            if len(selected_events) >= number:
                break
        return selected_events

    def query_events(self, selector: dict) -> dict[str, Event]:
        """Query the event server for events based on a selector. Requires an event server be configured."""
        events = OrderedDict()
        if self.event_server:
            response = requests.get(
                self.event_server + "/events",
                timeout=10,
                params={"selector": selector},
            )
            if not response.ok:
                response.raise_for_status()
            for key, value in response.json().items():
                events[key] = Event.model_validate(value)
            return dict(events)
        raise ValueError("No event server configured, cannot query events.")

    def log(self, event: Union[Event, Any], level: Optional[int] = None) -> None:
        """Log an event."""

        # * If we've got a string or dict, check if it's a serialized event
        if isinstance(event, str):
            with contextlib.suppress(ValidationError):
                event = Event.model_validate_json(event)
        if isinstance(event, dict):
            with contextlib.suppress(ValidationError):
                event = Event.model_validate(event)
        if not isinstance(event, Event):
            event = self._new_event_for_log(event, level)
        event.log_level = level if level else event.log_level
        event.source = event.source if event.source is not None else self.source
        self.logger.log(event.log_level, event.model_dump_json())
        # * Log the event to the event server if configured
        # * Only log if the event is at the same level or higher than the logger
        if self.logger.getEffectiveLevel() <= event.log_level:
            print(f"{event.event_timestamp} ({event.event_type}): {event.event_data}")
            if self.event_server:
                self._send_event_to_event_server(event)

    def log_debug(self, event: Union[Event, str]) -> None:
        """Log an event at the debug level."""
        self.log(event, logging.DEBUG)

    def log_info(self, event: Union[Event, str]) -> None:
        """Log an event at the info level."""
        self.log(event, logging.INFO)

    info = log_info

    def log_warning(self, event: Union[Event, str]) -> None:
        """Log an event at the warning level."""
        self.log(event, logging.WARNING)

    warning = log_warning

    def log_error(self, event: Union[Event, str]) -> None:
        """Log an event at the error level."""
        self.log(event, logging.ERROR)

    error = log_error

    def log_critical(self, event: Union[Event, str]) -> None:
        """Log an event at the critical level."""
        self.log(event, logging.CRITICAL)

    critical = log_critical

    @threaded_task
    def _send_event_to_event_server(self, event: Event) -> None:
        """Send an event to the event manager."""
        response = requests.post(
            url=self.event_server + "/event",
            json=event.model_dump(mode="json"),
            timeout=10,
        )
        if not response.ok:
            response.raise_for_status()

    def _new_event_for_log(self, event_data: Any, level: int) -> Event:
        """Create a new log event from arbitrary data"""
        event_type = EventType.LOG
        if level == logging.DEBUG:
            event_type = EventType.LOG_DEBUG
        elif level == logging.INFO:
            event_type = EventType.LOG_INFO
        elif level == logging.WARNING:
            event_type = EventType.LOG_WARNING
        elif level == logging.ERROR:
            event_type = EventType.LOG_ERROR
        elif level == logging.CRITICAL:
            event_type = EventType.LOG_CRITICAL
        return Event(
            event_type=event_type,
            event_data=event_data,
        )


default_logger = EventClient(
    config=EventClientConfig(name="madsci_default_log", log_level=logging.INFO)
)



"""Command Line Interface for managing MADSci Nodes."""

import contextlib
import os
from pathlib import Path
from typing import Optional

import click
from click.core import Context
from madsci.common.types.node_types import NodeDefinition, NodeType
from madsci.common.types.workcell_types import WorkcellDefinition
from madsci.common.utils import (
    PathLike,
    prompt_for_input,
    prompt_from_list,
    prompt_yes_no,
    relative_path,
    save_model,
    search_for_file_pattern,
    to_snake_case,
)
from rich.console import Console
from rich.pretty import pprint

console = Console()


class NodeContext:
    """Context object for node commands."""

    def __init__(self) -> None:
        """Initialize the context object."""
        self.node_def: Optional[NodeDefinition] = None
        self.path: Optional[Path] = None
        self.workcell_def: Optional[WorkcellDefinition] = None
        self.quiet: bool = False


pass_node = click.make_pass_decorator(NodeContext)


def find_node_template(name: Optional[str], path: Optional[str]) -> NodeContext:
    """Find a node template by name or path."""
    node_context = NodeContext()

    if path:
        node_context.path = Path(path)
        if node_context.path.exists():
            if path.endswith(".node.template.yaml"):
                node_context.node_def = NodeDefinition.from_yaml(path)
            return node_context

    node_template_files = search_for_file_pattern("*.node.template.yaml")
    for node_template_file in node_template_files:
        with contextlib.suppress(Exception):
            node_def = NodeDefinition.from_yaml(node_template_file)
            if not name or node_def.node_name == name:
                node_context.path = Path(node_template_file)
                node_context.node_def = node_def
                return node_context

    return node_context


def find_node(name: Optional[str], path: Optional[str]) -> NodeContext:
    """Find a node by name or path, including within workcell files."""
    node_context = NodeContext()

    if path:
        node_context.path = Path(path)
        if node_context.path.exists():
            if path.endswith(".node.yaml"):
                node_context.node_def = NodeDefinition.from_yaml(path)
            elif path.endswith(".workcell.yaml"):
                node_context.workcell_def = WorkcellDefinition.from_yaml(path)
                node_context.node_def = find_node_in_workcell(
                    name,
                    node_context.workcell_def,
                )
            return node_context

    node_files = search_for_file_pattern("*.node.yaml")
    for node_file in node_files:
        with contextlib.suppress(Exception):
            node_def = NodeDefinition.from_yaml(node_file)
            if not name or node_def.node_name == name:
                node_context.path = Path(node_file)
                node_context.node_def = node_def
                return node_context

    workcell_files = search_for_file_pattern("*.workcell.yaml")
    for workcell_file in workcell_files:
        with contextlib.suppress(Exception):
            workcell_def = WorkcellDefinition.from_yaml(workcell_file)
            node_def = find_node_in_workcell(name, workcell_def)
            if node_def:
                node_context.path = Path(workcell_file)
                node_context.workcell_def = workcell_def
                node_context.node_def = node_def
                return node_context

    return node_context


def find_node_in_workcell(
    name: Optional[str],
    workcell_def: WorkcellDefinition,
) -> Optional[NodeDefinition]:
    """Find a node definition within a workcell."""
    for node in workcell_def.nodes:
        if isinstance(node, NodeDefinition) and (not name or node.node_name == name):
            return node
        if isinstance(node, str) and node.endswith(".node.yaml"):
            node_def = NodeDefinition.from_yaml(node)
            if not name or node_def.node_name == name:
                return node_def
    return None


@click.group()
@click.option("--name", "-n", type=str, help="The name of the node to operate on.")
@click.option(
    "--path",
    "-p",
    type=str,
    help="The path to the node or workcell definition file.",
)
@click.pass_context
def node(ctx: Context, name: Optional[str], path: Optional[str]) -> None:
    """Manage nodes."""
    ctx.obj = find_node(name, path)
    ctx.obj.quiet = ctx.parent.params.get("quiet")


@node.command()
@click.option("--name", "-n", type=str, help="The name of the node.", required=False)
@click.option(
    "--path",
    "-p",
    type=str,
    help="The path to save the node definition file.",
)
@click.option("--description", "-d", type=str, help="The description of the node.")
@click.option(
    "--template_name",
    type=str,
    help="The name of the node template to use for the node.",
)
@click.option(
    "--template_path",
    type=str,
    help="Path to the node definition template file to use for the node.",
)
@click.option(
    "--standalone",
    "-s",
    is_flag=True,
    help="Don't add node to any workcell.",
)
@click.option(
    "--is_template",
    "-t",
    is_flag=True,
    help="Create a node template.",
)
@click.pass_context
def create(
    ctx: Context,
    name: Optional[str],
    path: Optional[str],
    description: Optional[str],
    template_name: Optional[str],
    template_path: Optional[str],
    standalone: bool,
    is_template: bool,
) -> None:
    """Create a new node."""
    name = name if name else ctx.parent.params.get("name")
    name = (
        name
        if name
        else prompt_for_input("Node Name", required=True, quiet=ctx.obj.quiet)
    )
    description = (
        description
        if description
        else prompt_for_input("Node Description", quiet=ctx.obj.quiet)
    )
    template_definition = get_template_definition(ctx, template_name, template_path)
    if not template_definition and (template_name or template_path):
        return

    if not template_definition:
        module_name = prompt_for_input(
            "Module Name",
            required=True,
            quiet=ctx.obj.quiet,
        )
        node_type = prompt_from_list(
            "Node Type",
            required=False,
            options=[node_type.value for node_type in NodeType],
            default=None,
            quiet=ctx.obj.quiet,
        )
        node_definition = NodeDefinition(
            node_name=name,
            node_description=description,
            is_template=is_template,
            module_name=module_name,
            node_type=node_type,
        )
    else:
        node_definition = NodeDefinition(
            **template_definition.model_dump(exclude={"node_id", "is_template"}),
            node_name=name,
            node_description=description,
            is_template=is_template,
        )
    console.print(node_definition)

    path = path if path else ctx.parent.params.get("path")
    if not path:
        if Path.cwd().name == "nodes":
            default_path = Path.cwd() / f"{to_snake_case(name)}.node.yaml"
        else:
            default_path = Path.cwd() / "nodes" / f"{to_snake_case(name)}.node.yaml"
        new_path = prompt_for_input(
            "Path to save Node Definition file",
            default=str(default_path),
            quiet=ctx.obj.quiet,
        )
        if new_path:
            path = Path(new_path)
        if not path.expanduser().parent.exists():
            console.print(f"Creating directory: {path.parent}")
            path.expanduser().parent.mkdir(parents=True, exist_ok=True)
    save_model(path=path, model=node_definition, overwrite_check=not ctx.obj.quiet)

    # *Handle workcell integration
    if (
        not standalone
        and not is_template
        and prompt_yes_no(
            "Add node to a workcell?",
            default=True,
            quiet=ctx.obj.quiet,
        )
    ):
        add_node_to_workcell(ctx, name, path, node_definition)


def get_template_definition(
    ctx: Context, template_name: Optional[str], template_path: Optional[str]
) -> Optional[NodeDefinition]:
    """Get a node template definition, if applicable."""
    template_definition = None
    if template_name or template_path:
        node_template = find_node_template(template_name, template_path)
        if node_template.node_def:
            template_definition = node_template.node_def
        else:
            console.print(
                f"Node template not found: {template_name or template_path}",
            )
    else:
        node_templates = search_for_file_pattern("*.node.template.yaml")
        if node_templates:
            template_path = prompt_from_list(
                prompt="Node Definition Template Files",
                options=node_templates,
                default=None,
                required=False,
                quiet=ctx.obj.quiet,
            )
            try:
                template_definition = NodeDefinition.from_yaml(template_path)
            except Exception as e:
                console.print(f"Error creating node from template: {e}")
    return template_definition


def add_node_to_workcell(
    ctx: Context,
    name: str,
    path: PathLike,
    node_definition: NodeDefinition,
) -> None:
    """Adds a node definition to a workcell definition's 'nodes' section"""
    workcell_files = search_for_file_pattern("*.workcell.yaml")
    if workcell_files:
        if ctx.obj.quiet:
            # *In quiet mode, automatically add to first workcell
            workcell_path = workcell_files[0]
        else:
            workcell_path = prompt_from_list(
                prompt="Add node to workcell",
                options=workcell_files,
                default=workcell_files[0],
            )

        if workcell_path:
            workcell_def = WorkcellDefinition.from_yaml(workcell_path)
            # *Calculate relative path from workcell to node
            workcell_dir = Path(workcell_path).parent
            rel_path = str(
                relative_path(target=path.absolute(), source=workcell_dir.absolute()),
            )

            # *Add node to workcell if not already present
            workcell_def.nodes[node_definition.node_name] = rel_path
            save_model(workcell_path, workcell_def, overwrite_check=False)
            console.print(
                f"Added node [bold]{name}[/] to workcell: [bold]{workcell_path}[/]",
            )


@node.command()
def list() -> None:
    """List all nodes, including those in workcell files."""
    node_files = search_for_file_pattern("*.node.yaml")
    workcell_files = search_for_file_pattern("*.workcell.yaml")

    nodes_found = False

    if node_files:
        for node_file in sorted(set(node_files)):
            node_definition = NodeDefinition.from_yaml(node_file)
            console.print(
                f"[bold]{node_definition.node_name}[/]: {node_definition.node_description} ({node_file})",
            )
            nodes_found = True

    for workcell_file in workcell_files:
        workcell_def = WorkcellDefinition.from_yaml(workcell_file)
        for node in workcell_def.nodes:
            if isinstance(node, NodeDefinition):
                console.print(
                    f"[bold]{node.node_name}[/]: {node.node_description} (in {workcell_file})",
                )
                nodes_found = True

    if not nodes_found:
        console.print("No node definitions found")


@node.command()
@pass_node
def info(ctx: NodeContext) -> None:
    """Get information about a node."""
    if ctx.node_def:
        pprint(ctx.node_def)
    else:
        console.print(
            "No node found. Specify node by name or path. If you don't have a node file, you can create one with 'madsci node create'.",
        )


@node.command()
@click.option("--yes", "-y", is_flag=True, help="Skip confirmation prompt.")
@pass_node
def delete(ctx: NodeContext, yes: bool) -> None:
    """Delete a node."""
    if ctx.node_def and ctx.path:
        console.print(f"Deleting node: {ctx.node_def.node_name} ({ctx.path})")
        if yes or ctx.quiet or prompt_yes_no("Are you sure?"):
            # First find all workcells that contain this node
            workcell_files = search_for_file_pattern("*.workcell.yaml")
            for workcell_file in workcell_files:
                workcell_def = WorkcellDefinition.from_yaml(workcell_file)
                if ctx.node_def.node_name in workcell_def.nodes and (
                    yes
                    or ctx.quiet
                    or prompt_yes_no(
                        f"Remove from workcell [bold]{workcell_def.name}[/] ([italic]{workcell_file}[/])?",
                        default=True,
                    )
                ):
                    del workcell_def.nodes[ctx.node_def.node_name]
                    save_model(workcell_file, workcell_def, overwrite_check=False)
                    console.print(f"Removed from workcell: {workcell_file}")

            # Finally delete the node file
            ctx.path.unlink()
            console.print(f"Deleted {ctx.path}")
    else:
        console.print(
            "No node found. Specify node by name or path. If you don't have a node file, you can create one with 'madsci node create'.",
        )


@node.command()
@pass_node
def validate(ctx: NodeContext) -> None:
    """Validate a node definition file."""
    if ctx.node_def:
        console.print(ctx.node_def)
    else:
        console.print(
            "No node found. Specify node by name or path. If you don't have a node definition file, you can create one with 'madsci node create'.",
        )


@node.command()
@click.option("--command_name", "--name", "-n", type=str, required=False)
@click.option("--command", "-c", type=str, required=False)
@pass_node
def add_command(ctx: NodeContext, command_name: str, command: str) -> None:
    """Add a command to a node definition."""
    if not ctx.node_def:
        console.print(
            "No node found. Specify node by name or path. If you don't have a node file, you can create one with 'madsci node create'.",
        )
        return

    if not command_name:
        command_name = prompt_for_input(
            "Command Name",
            required=True,
            quiet=ctx.obj.quiet,
        )
    if not command:
        command = prompt_for_input("Command", required=True, quiet=ctx.obj.quiet)

    if command_name in ctx.node_def.commands:
        console.print(
            f"Command [bold]{command_name}[/] already exists in node definition: [bold]{ctx.node_def.node_name}[/] ({ctx.path})",
        )
        if not prompt_yes_no(
            "Do you want to overwrite it?",
            default="no",
            quiet=ctx.quiet,
        ):
            return

    ctx.node_def.commands[command_name] = command
    save_model(ctx.path, ctx.node_def, overwrite_check=False)
    console.print(
        f"Added command [bold]{command_name}[/] to node: [bold]{ctx.node_def.node_name}[/]",
    )


@node.command()
@click.argument("command_name", type=str)
@pass_node
def run(ctx: NodeContext, command_name: str) -> None:
    """Run a command in a node."""
    if not ctx.node_def:
        console.print(
            "No node found. Specify node by name or path. If you don't have a node file, you can create one with 'madsci node create'.",
        )
        return

    if command_name in ctx.node_def.commands:
        command = ctx.node_def.commands[command_name]
        console.print(
            f"Running command: [bold]{command_name}[/] ({command}) in node: [bold]{ctx.node_def.node_name}[/] ({ctx.path})",
        )
        print(os.popen(command).read())  # noqa: S605,T201
    else:
        console.print(
            f"Command [bold]{command_name}[/] not found in node definition: [bold]{ctx.node_def.node_name}[/] ({ctx.path})",
        )


@node.command()
@click.argument("command_name", type=str, required=False)
@click.option("--yes", "-y", is_flag=True, help="Skip confirmation prompt.")
@pass_node
def delete_command(ctx: NodeContext, command_name: str, yes: bool) -> None:
    """Delete a command from a node definition."""
    if not ctx.node_def:
        console.print(
            "No node found. Specify node by name or path. If you don't have a node file, you can create one with 'madsci node create'.",
        )
        return

    if not command_name:
        if not ctx.node_def.commands:
            console.print("No commands found in node definition.")
            return
        command_name = prompt_from_list(
            "Select command to delete",
            options=list(ctx.node_def.commands.keys()),
            required=True,
            quiet=ctx.obj.quiet,
        )

    if command_name in ctx.node_def.commands:
        if (
            yes
            or ctx.quiet
            or prompt_yes_no(
                f"Are you sure you want to delete command [bold]{command_name}[/]?",
                default="no",
                quiet=ctx.obj.quiet,
            )
        ):
            del ctx.node_def.commands[command_name]
            save_model(ctx.path, ctx.node_def, overwrite_check=False)
            console.print(
                f"Deleted command [bold]{command_name}[/] from node: [bold]{ctx.node_def.node_name}[/]",
            )
    else:
        console.print(
            f"Command [bold]{command_name}[/] not found in node definition: [bold]{ctx.node_def.node_name}[/] ({ctx.path})",
        )



"""Fast API Client for Resources"""

import time
from datetime import datetime
from typing import Any, Optional, Union

import requests
from madsci.client.event_client import EventClient
from madsci.common.types.auth_types import OwnershipInfo
from madsci.common.types.resource_types import (
    GridIndex2D,
    GridIndex3D,
    Resource,
    ResourceDataModels,
)
from madsci.common.types.resource_types.definitions import ResourceDefinitions
from madsci.common.types.resource_types.server_types import (
    PushResourceBody,
    RemoveChildBody,
    ResourceGetQuery,
    ResourceHistoryGetQuery,
    SetChildBody,
)


class ResourceClient:
    """REST client for interacting with a MADSci Resource Manager."""

    local_resources: dict[str, ResourceDataModels]

    def __init__(
        self,
        url: Optional[str] = None,
        event_client: Optional[EventClient] = None,
        ownership_info: Optional[OwnershipInfo] = None,
    ) -> None:
        """Initialize the resource client."""
        self.url = str(url) if url is not None else None
        self.ownership_info = ownership_info if ownership_info else OwnershipInfo()
        if self.url is not None and str(self.url).endswith("/"):
            self.url = str(self.url)[:-1]
        if self.url is not None:
            start_time = time.time()
            while time.time() - start_time < 20:
                try:
                    requests.get(self.url + "/definition", timeout=10)
                    break
                except Exception:
                    time.sleep(1)
            else:
                raise ConnectionError(
                    f"Could not connect to the resource manager at {self.url}."
                )
        self.local_resources = {}
        self.logger = event_client if event_client is not None else EventClient()
        if self.url is None:
            self.logger.log_warning(
                "ResourceClient initialized without a URL. Resource operations will be local-only and won't be persisted to a server. Local-only mode has limited functionality and should be used only for basic development purposes only. DO NOT USE LOCAL-ONLY MODE FOR PRODUCTION."
            )

    def add_resource(self, resource: Resource) -> Resource:
        """
        Add a resource to the server.

        Args:
            resource (Resource): The resource to add.

        Returns:
            Resource: The added resource as returned by the server.
        """
        if self.url:
            response = requests.post(
                f"{self.url}/resource/add",
                json=resource.model_dump(mode="json"),
                timeout=10,
            )
            response.raise_for_status()
            resource = Resource.discriminate(response.json())
            resource.resource_url = f"{self.url}/resource/{resource.resource_id}"
        else:
            self.local_resources[resource.resource_id] = resource
        return resource

    def init_resource(
        self, resource_definition: ResourceDefinitions
    ) -> ResourceDataModels:
        """
        Initializes a resource with the resource manager based on a definition, either creating a new resource if no matching one exists, or returning an existing match.

        Args:
            resource (Resource): The resource to initialize.

        Returns:
            ResourceDataModels: The initialized resource as returned by the server.
        """
        if self.url:
            response = requests.post(
                f"{self.url}/resource/init",
                json=resource_definition.model_dump(mode="json"),
                timeout=10,
            )
            response.raise_for_status()

            resource = Resource.discriminate(response.json())
            resource.resource_url = f"{self.url}/resource/{resource.resource_id}"
        else:
            self.logger.log_warning(
                "Local-only mode does not check to see if an existing resource match already exists."
            )
            resource = Resource.discriminate(resource_definition)
            self.local_resources[resource.resource_id] = resource
        return resource

    def add_or_update_resource(self, resource: Resource) -> Resource:
        """
        Add a resource to the server.

        Args:
            resource (Resource): The resource to add.

        Returns:
            Resource: The added resource as returned by the server.
        """
        if self.url:
            response = requests.post(
                f"{self.url}/resource/add_or_update",
                json=resource.model_dump(mode="json"),
                timeout=10,
            )
            response.raise_for_status()
            resource = Resource.discriminate(response.json())
            resource.resource_url = f"{self.url}/resource/{resource.resource_id}"
        else:
            self.local_resources[resource.resource_id] = resource
        return resource

    def update_resource(self, resource: ResourceDataModels) -> ResourceDataModels:
        """
        Update or refresh a resource, including its children, on the server.

        Args:
            resource (ResourceDataModels): The resource to update.

        Returns:
            ResourceDataModels: The updated resource as returned by the server.
        """
        if self.url:
            response = requests.post(
                f"{self.url}/resource/update",
                json=resource.model_dump(mode="json"),
                timeout=10,
            )
            response.raise_for_status()
            resource = Resource.discriminate(response.json())
            resource.resource_url = f"{self.url}/resource/{resource.resource_id}"
        else:
            self.local_resources[resource.resource_id] = resource
        return resource

    def get_resource(
        self,
        resource_id: str,
    ) -> ResourceDataModels:
        """
        Retrieve a resource from the server.

        Args:
            resource_id (str): The ID of the resource to retrieve.

        Returns:
            ResourceDataModels: The retrieved resource.
        """
        if self.url:
            response = requests.get(
                f"{self.url}/resource/{resource_id}",
                timeout=10,
            )
            response.raise_for_status()
            resource = Resource.discriminate(response.json())
            resource.resource_url = f"{self.url}/resource/{resource.resource_id}"
        else:
            self.logger.log_warning(
                "Local-only mode does not currently search through child resources to get children."
            )
            resource = self.local_resources.get(resource_id)
        return resource

    def query_resource(
        self,
        resource_id: Optional[str] = None,
        resource_name: Optional[str] = None,
        parent_id: Optional[str] = None,
        resource_type: Optional[str] = None,
        base_type: Optional[str] = None,
        unique: Optional[bool] = False,
        multiple: Optional[bool] = False,
    ) -> Union[ResourceDataModels, list[ResourceDataModels]]:
        """
        Query for one or more resources matching specific properties.

        Args:
            resource_id (str): The ID of the resource to retrieve.
            resource_name (str): The name of the resource to retrieve.
            parent_id (str): The ID of the parent resource.
            resource_type (str): The custom type of the resource.
            base_type (str): The base type of the resource.
            unique (bool): Whether to require a unique resource or not.
            multiple (bool): Whether to return multiple resources or just the first.

        Returns:
            Resource: The retrieved resource.
        """
        if self.url:
            payload = ResourceGetQuery(
                resource_id=resource_id,
                resource_name=resource_name,
                parent_id=parent_id,
                resource_type=resource_type,
                base_type=base_type,
                unique=unique,
                multiple=multiple,
            ).model_dump(mode="json")
            response = requests.post(
                f"{self.url}/resource/query", json=payload, timeout=10
            )
            response.raise_for_status()
            response_json = response.json()
            if isinstance(response_json, list):
                resources = [
                    Resource.discriminate(resource) for resource in response_json
                ]
                for resource in resources:
                    resource.resource_url = self.url
                return resources
            resource = Resource.discriminate(response.json())
            resource.resource_url = f"{self.url}/resource/{resource.resource_id}"
        else:
            self.logger.log_error(
                "Local-only mode does not currently support querying."
            )
            raise NotImplementedError(
                "Local-only mode does not currently support querying."
            )
        return resource

    def remove_resource(self, resource_id: str) -> dict[str, Any]:
        """
        Remove a resource by moving it to the history table with `removed=True`.
        """
        if self.url:
            response = requests.delete(f"{self.url}/resource/{resource_id}", timeout=10)
            response.raise_for_status()
            resource = Resource.discriminate(response.json())
            resource.resource_url = f"{self.url}/resource/{resource.resource_id}"
        else:
            resource = self.local_resources.pop(resource_id)
            resource.removed = True
        return resource

    def query_history(
        self,
        resource_id: Optional[str] = None,
        version: Optional[int] = None,
        change_type: Optional[str] = None,
        removed: Optional[bool] = None,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        limit: Optional[int] = 100,
    ) -> list[dict[str, Any]]:
        """
        Retrieve the history of a resource with flexible filters.
        """
        if self.url:
            query = ResourceHistoryGetQuery(
                resource_id=resource_id,
                version=version,
                change_type=change_type,
                removed=removed,
                start_date=start_date,
                end_date=end_date,
                limit=limit,
            ).model_dump(mode="json")
            response = requests.post(
                f"{self.url}/history/query", json=query, timeout=10
            )
            response.raise_for_status()
        else:
            self.logger.log_error(
                "Local-only mode does not currently support querying history."
            )
            raise NotImplementedError(
                "Local-only mode does not currently support querying history."
            )

        return response.json()

    def restore_deleted_resource(self, resource_id: str) -> dict[str, Any]:
        """
        Restore a deleted resource from the history table.
        """
        if self.url:
            response = requests.post(
                f"{self.url}/history/{resource_id}/restore", timeout=10
            )
            response.raise_for_status()
            resource = Resource.discriminate(response.json())
            resource.resource_url = f"{self.url}/resource/{resource.resource_id}"
        else:
            self.logger.log_error(
                "Local-only mode does not currently support restoring resources."
            )
            raise NotImplementedError(
                "Local-only mode does not currently support restoring resources."
            )
        return resource

    def push(
        self,
        resource: Union[ResourceDataModels, str],
        child: Union[ResourceDataModels, str],
    ) -> ResourceDataModels:
        """
        Push a child resource onto a parent stack or queue.

        Args:
            resource (Union[ResourceDataModels, str]): The parent resource or its ID.
            child (Union[ResourceDataModels, str]): The child resource or its ID.

        Returns:
            ResourceDataModels: The updated parent resource.
        """
        if self.url:
            resource_id = (
                resource.resource_id if isinstance(resource, Resource) else resource
            )
            payload = PushResourceBody(
                child=child if isinstance(child, Resource) else None,
                child_id=child.resource_id if isinstance(child, Resource) else child,
            ).model_dump(mode="json")
            response = requests.post(
                f"{self.url}/resource/{resource_id}/push", json=payload, timeout=10
            )
            response.raise_for_status()
            resource = Resource.discriminate(response.json())
            resource.resource_url = f"{self.url}/resource/{resource.resource_id}"
        else:
            resource = resource.children.append(child)
            self.local_resources[resource.resource_id] = resource
        return resource

    def pop(
        self, resource: Union[str, ResourceDataModels]
    ) -> tuple[ResourceDataModels, ResourceDataModels]:
        """
        Pop an asset from a stack or queue resource.

        Args:
            resource (Union[str, ResourceDataModels]): The parent resource or its ID.

        Returns:
            tuple[ResourceDataModels, ResourceDataModels]: The popped asset and updated parent.
        """

        if self.url:
            resource_id = (
                resource.resource_id if isinstance(resource, Resource) else resource
            )
            response = requests.post(
                f"{self.url}/resource/{resource_id}/pop", timeout=10
            )
            response.raise_for_status()
            result = response.json()
            popped_asset = Resource.discriminate(result[0])
            update_parent = Resource.discriminate(result[1])
            popped_asset.resource_url = (
                f"{self.url}/resource/{popped_asset.resource_id}"
            )
            update_parent.resource_url = (
                f"{self.url}/resource/{update_parent.resource_id}"
            )
        else:
            update_parent = resource
            popped_asset = update_parent.children.pop(0)
            self.local_resources[update_parent.resource_id] = update_parent
        return popped_asset, update_parent

    def set_child(
        self,
        resource: Union[str, ResourceDataModels],
        key: Union[str, GridIndex2D, GridIndex3D],
        child: Union[str, ResourceDataModels],
    ) -> ResourceDataModels:
        """
        Set a child resource in a parent container resource.

        Args:
            resource (Union[str, ResourceDataModels]): The parent container resource or its ID.
            key (Union[str, GridIndex2D, GridIndex3D]): The key to identify the child resource's location in the parent container.
            child (Union[str, ResourceDataModels]): The child resource or its ID.

        Returns:
            ResourceDataModels: The updated parent container resource.
        """
        if self.url:
            resource_id = (
                resource.resource_id if isinstance(resource, Resource) else resource
            )
            payload = SetChildBody(
                key=key,
                child=child,
            ).model_dump(mode="json")
            response = requests.post(
                f"{self.url}/resource/{resource_id}/child/set",
                json=payload,
                timeout=10,
            )
            response.raise_for_status()
            resource = Resource.discriminate(response.json())
            resource.resource_url = f"{self.url}/resource/{resource.resource_id}"
        else:
            resource = resource.children[key] = child
            self.local_resources[resource.resource_id] = resource
        return resource

    def remove_child(
        self,
        resource: Union[str, ResourceDataModels],
        key: Union[str, GridIndex2D, GridIndex3D],
    ) -> ResourceDataModels:
        """
        Remove a child resource from a parent container resource.

        Args:
            resource (Union[str, ResourceDataModels]): The parent container resource or its ID.
            key (Union[str, GridIndex2D, GridIndex3D]): The key to identify the child resource's location in the parent container.

        Returns:
            ResourceDataModels: The updated parent container resource.
        """
        if self.url:
            resource_id = (
                resource.resource_id if isinstance(resource, Resource) else resource
            )
            payload = RemoveChildBody(
                key=key,
            ).model_dump(mode="json")
            response = requests.post(
                f"{self.url}/resource/{resource_id}/child/remove",
                json=payload,
                timeout=10,
            )
            response.raise_for_status()
            resource = Resource.discriminate(response.json())
            resource.resource_url = f"{self.url}/resource/{resource.resource_id}"
        else:
            resource = resource.children.pop(key)
            self.local_resources[resource.resource_id] = resource
        return resource

    def set_quantity(
        self, resource: Union[str, ResourceDataModels], quantity: Union[float, int]
    ) -> ResourceDataModels:
        """
        Set the quantity of a resource.

        Args:
            resource (Union[str, ResourceDataModels]): The resource or its ID.
            quantity (Union[float, int]): The quantity to set.

        Returns:
            ResourceDataModels: The updated resource.
        """
        if self.url:
            resource_id = (
                resource.resource_id if isinstance(resource, Resource) else resource
            )
            response = requests.post(
                f"{self.url}/resource/{resource_id}/quantity",
                params={"quantity": quantity},
                timeout=10,
            )
            response.raise_for_status()
            resource = Resource.discriminate(response.json())
            resource.resource_url = f"{self.url}/resource/{resource.resource_id}"
        else:
            self.local_resources[resource.resource_id].quantity = quantity
        return resource

    def change_quantity_by(
        self, resource: Union[str, ResourceDataModels], amount: Union[float, int]
    ) -> ResourceDataModels:
        """
        Change the quantity of a resource by a given amount.

        Args:
            resource (Union[str, ResourceDataModels]): The resource or its ID.
            amount (Union[float, int]): The quantity to change by.

        Returns:
            ResourceDataModels: The updated resource.
        """
        if self.url:
            resource_id = (
                resource.resource_id if isinstance(resource, Resource) else resource
            )
            response = requests.post(
                f"{self.url}/resource/{resource_id}/quantity/change_by",
                params={"amount": amount},
                timeout=10,
            )
            response.raise_for_status()
            resource = Resource.discriminate(response.json())
            resource.resource_url = f"{self.url}/resource/{resource.resource_id}"
        else:
            self.local_resources[resource.resource_id].quantity += amount
            resource = self.local_resources[resource.resource_id]
        return resource

    def increase_quantity(
        self, resource: Union[str, ResourceDataModels], amount: Union[float, int]
    ) -> ResourceDataModels:
        """
        Increase the quantity of a resource by a given amount.

        Args:
            resource (Union[str, ResourceDataModels]): The resource or its ID.
            amount (Union[float, int]): The quantity to increase by. Note that this is a magnitude, so negative and positive values will have the same effect.

        Returns:
            ResourceDataModels: The updated resource.
        """
        if self.url:
            resource_id = (
                resource.resource_id if isinstance(resource, Resource) else resource
            )
            response = requests.post(
                f"{self.url}/resource/{resource_id}/quantity/increase",
                params={"amount": amount},
                timeout=10,
            )
            response.raise_for_status()
            resource = Resource.discriminate(response.json())
            resource.resource_url = f"{self.url}/resource/{resource.resource_id}"
        else:
            self.local_resources[resource.resource_id].quantity += abs(amount)
            resource = self.local_resources[resource.resource_id]
        return resource

    def decrease_quantity(
        self, resource: Union[str, ResourceDataModels], amount: Union[float, int]
    ) -> ResourceDataModels:
        """
        Decrease the quantity of a resource by a given amount.

        Args:
            resource (Union[str, ResourceDataModels]): The resource or its ID.
            amount (Union[float, int]): The quantity to decrease by. Note that this is a magnitude, so negative and positive values will have the same effect.

        Returns:
            ResourceDataModels: The updated resource.
        """
        if self.url:
            resource_id = (
                resource.resource_id if isinstance(resource, Resource) else resource
            )
            response = requests.post(
                f"{self.url}/resource/{resource_id}/quantity/decrease",
                params={"amount": amount},
                timeout=10,
            )
            response.raise_for_status()
            resource = Resource.discriminate(response.json())
            resource.resource_url = f"{self.url}/resource/{resource.resource_id}"
        else:
            self.local_resources[resource.resource_id].quantity -= abs(amount)
            resource = self.local_resources[resource.resource_id]
        return resource

    def set_capacity(
        self, resource: Union[str, ResourceDataModels], capacity: Union[float, int]
    ) -> ResourceDataModels:
        """
        Set the capacity of a resource.

        Args:
            resource (Union[str, ResourceDataModels]): The resource or its ID.
            capacity (Union[float, int]): The capacity to set.

        Returns:
            ResourceDataModels: The updated resource.
        """
        if self.url:
            resource_id = (
                resource.resource_id if isinstance(resource, Resource) else resource
            )
            response = requests.post(
                f"{self.url}/resource/{resource_id}/capacity",
                params={"capacity": capacity},
                timeout=10,
            )
            response.raise_for_status()
            resource = Resource.discriminate(response.json())
            resource.resource_url = f"{self.url}/resource/{resource.resource_id}"
        else:
            self.local_resources[resource.resource_id].capacity = capacity
            resource = self.local_resources[resource.resource_id]
        return resource

    def remove_capacity_limit(
        self, resource: Union[str, ResourceDataModels]
    ) -> ResourceDataModels:
        """
        Remove the capacity limit of a resource.

        Args:
            resource (Union[str, ResourceDataModels]): The resource or its ID.

        Returns:
            ResourceDataModels: The updated resource.
        """
        if self.url:
            resource_id = (
                resource.resource_id if isinstance(resource, Resource) else resource
            )
            response = requests.delete(
                f"{self.url}/resource/{resource_id}/capacity", timeout=10
            )
            response.raise_for_status()
            resource = Resource.discriminate(response.json())
            resource.resource_url = f"{self.url}/resource/{resource.resource_id}"
        else:
            self.local_resources[resource.resource_id].capacity = None
            resource = self.local_resources[resource.resource_id]
        return resource

    def empty(self, resource: Union[str, ResourceDataModels]) -> ResourceDataModels:
        """
        Empty the contents of a container or consumable resource.

        Args:
            resource (Union[str, ResourceDataModels]): The resource or its ID.

        Returns:
            ResourceDataModels: The updated resource.
        """
        if self.url:
            resource_id = (
                resource.resource_id if isinstance(resource, Resource) else resource
            )
            response = requests.post(
                f"{self.url}/resource/{resource_id}/empty",
                timeout=10,
            )
            response.raise_for_status()
            resource = Resource.discriminate(response.json())
            resource.resource_url = f"{self.url}/resource/{resource.resource_id}"
        else:
            resource = self.local_resources[resource.resource_id]
            while resource.children:
                resource.children.pop()
            if resource.quantity:
                resource.quantity = 0
            self.local_resources[resource.resource_id] = resource
        return resource

    def fill(self, resource: Union[str, ResourceDataModels]) -> ResourceDataModels:
        """
        Fill a consumable resource to capacity.

        Args:
            resource (Union[str, ResourceDataModels]): The resource or its ID.

        Returns:
            ResourceDataModels: The updated resource.
        """
        if self.url:
            resource_id = (
                resource.resource_id if isinstance(resource, Resource) else resource
            )
            response = requests.post(
                f"{self.url}/resource/{resource_id}/fill",
                timeout=10,
            )
            response.raise_for_status()
            resource = Resource.discriminate(response.json())
            resource.resource_url = f"{self.url}/resource/{resource.resource_id}"
        else:
            resource = self.local_resources[resource.resource_id]
            resource.quantity = resource.capacity
            self.local_resources[resource.resource_id] = resource
        return resource



"""Automated Tests for the MADSci cli's manager commands."""

import os
from pathlib import Path

import pytest
from click.testing import CliRunner
from madsci.client.cli import root_cli
from madsci.common.types.lab_types import ManagerDefinition, ManagerType


@pytest.fixture
def runner() -> CliRunner:
    """Fixture for creating a CliRunner instance."""
    return CliRunner()


@pytest.fixture
def temp_manager_file(tmp_path: Path) -> Path:
    """Fixture for creating a temporary manager definition file."""
    manager_file = tmp_path / "test_manager.manager.yaml"
    manager_definition = ManagerDefinition(
        name="test_manager",
        description="A test manager",
        manager_type=ManagerType.EVENT_MANAGER,
    )
    manager_definition.to_yaml(manager_file)
    return manager_file


def test_manager_add(runner: CliRunner, tmp_path: Path) -> None:
    """Test adding a new manager definition."""
    result = runner.invoke(
        root_cli,
        [
            "manager",
            "add",
            "--name",
            "test_manager",
            "--description",
            "A test manager",
            "--manager_type",
            "event_manager",
            "--path",
            str(tmp_path / "test_manager.manager.yaml"),
        ],
    )
    assert result.exit_code == 0
    assert (tmp_path / "test_manager.manager.yaml").exists()


def test_manager_list(runner: CliRunner, temp_manager_file: Path) -> None:
    """Test listing all manager definitions."""
    os.chdir(temp_manager_file.parent)
    result = runner.invoke(root_cli, ["manager", "list"])
    assert result.exit_code == 0
    assert "test_manager" in result.output


def test_manager_info(runner: CliRunner, temp_manager_file: Path) -> None:
    """Test getting information about a specific manager definition."""
    os.chdir(temp_manager_file.parent)
    result = runner.invoke(
        root_cli, ["manager", "--path", str(temp_manager_file), "info"]
    )
    assert result.exit_code == 0
    assert "test_manager" in result.output


def test_manager_delete(runner: CliRunner, temp_manager_file: Path) -> None:
    """Test deleting a manager definition."""
    os.chdir(temp_manager_file.parent)
    result = runner.invoke(
        root_cli, ["manager", "--path", str(temp_manager_file), "delete", "--yes"]
    )
    assert result.exit_code == 0
    assert not temp_manager_file.exists()


def test_manager_validate(runner: CliRunner, temp_manager_file: Path) -> None:
    """Test validating a manager definition file."""
    os.chdir(temp_manager_file.parent)
    result = runner.invoke(
        root_cli, ["manager", "--path", str(temp_manager_file), "validate"]
    )
    assert result.exit_code == 0
    assert "test_manager" in result.output



"""REST-based node client implementation."""

import json
import tempfile
import time
from pathlib import Path
from typing import Any, ClassVar, Optional
from zipfile import ZipFile

import requests
from madsci.client.event_client import EventClient
from madsci.client.node.abstract_node_client import (
    AbstractNodeClient,
)
from madsci.common.types.action_types import ActionRequest, ActionResult, ActionStatus
from madsci.common.types.admin_command_types import AdminCommandResponse
from madsci.common.types.event_types import Event, EventClientConfig
from madsci.common.types.node_types import (
    AdminCommands,
    NodeClientCapabilities,
    NodeInfo,
    NodeSetConfigResponse,
    NodeStatus,
)
from madsci.common.types.resource_types.definitions import ResourceDefinition
from pydantic import AnyUrl


class RestNodeClient(AbstractNodeClient):
    """REST-based node client."""

    url_protocols: ClassVar[list[str]] = ["http", "https"]
    """The protocols supported by this client."""

    supported_capabilities: NodeClientCapabilities = NodeClientCapabilities(
        # *Supported capabilities
        get_info=True,
        get_state=True,
        get_status=True,
        send_action=True,
        get_action_result=True,
        get_action_history=True,
        action_files=True,
        send_admin_commands=True,
        set_config=True,
        get_log=True,
        # *Unsupported Capabilities
        get_resources=False,
    )

    def __init__(
        self, url: AnyUrl, event_client_config: Optional[EventClientConfig] = None
    ) -> "RestNodeClient":
        """Initialize the client."""
        super().__init__(url)
        self.logger = EventClient(config=event_client_config)

    def send_action(
        self,
        action_request: ActionRequest,
        await_result: bool = True,
        timeout: Optional[float] = None,
    ) -> ActionResult:
        """Perform an action on the node."""
        files = []
        try:
            files = [
                ("files", (file, Path(path).expanduser().open("rb")))
                for file, path in action_request.files.items()
            ]

            serialized_args = action_request.model_dump(mode="json")["args"]

            rest_response = requests.post(
                f"{self.url}/action",
                params={
                    "action_name": action_request.action_name,
                    "args": json.dumps(serialized_args),
                    "action_id": action_request.action_id,
                },
                files=files,
                timeout=60,
            )
        finally:
            # * Ensure files are closed
            for file in files:
                file[1][1].close()
        try:
            rest_response.raise_for_status()
        except requests.HTTPError as e:
            self.logger.log_error(f"{rest_response.status_code}: {rest_response.text}")
            raise e
        if "x-madsci-status" in rest_response.headers:
            response = process_file_response(rest_response)
        else:
            response = ActionResult.model_validate(rest_response.json())
        if await_result and not response.status.is_terminal:
            response = self.await_action_result(response.action_id, timeout=timeout)
        return response

    def get_action_history(
        self, action_id: Optional[str] = None
    ) -> dict[str, list[ActionResult]]:
        """Get the history of a single action performed on the node, or every action, if no action_id is specified."""
        response = requests.get(
            f"{self.url}/action", params={"action_id": action_id}, timeout=10
        )
        response.raise_for_status()
        return response.json()

    def get_action_result(self, action_id: str) -> ActionResult:
        """Get the result of an action on the node."""
        rest_response = requests.get(
            f"{self.url}/action/{action_id}",
            timeout=10,
        )
        rest_response.raise_for_status()
        if "x-madsci-status" in rest_response.headers:
            response = process_file_response(rest_response)
        else:
            response = ActionResult.model_validate(rest_response.json())
        return response

    def await_action_result(
        self, action_id: str, timeout: Optional[float] = None
    ) -> ActionResult:
        """Wait for an action to complete and return the result. Optionally, specify a timeout in seconds."""
        start_time = time.time()
        interval = 0.25
        while True:
            if timeout is not None and time.time() - start_time > timeout:
                raise TimeoutError("Timed out waiting for action to complete.")
            response = self.get_action_result(action_id)
            if not response.status.is_terminal:
                time.sleep(interval)
                interval = (
                    interval * 1.5 if interval < 5 else 5
                )  # * Capped Exponential backoff
                continue
            return response

    def get_status(self) -> NodeStatus:
        """Get the status of the node."""
        response = requests.get(f"{self.url}/status", timeout=10)
        response.raise_for_status()
        return NodeStatus.model_validate(response.json())

    def get_state(self) -> dict[str, Any]:
        """Get the state of the node."""
        response = requests.get(f"{self.url}/state", timeout=10)
        response.raise_for_status()
        return response.json()

    def get_info(self) -> NodeInfo:
        """Get information about the node."""
        response = requests.get(f"{self.url}/info", timeout=10)
        response.raise_for_status()
        return NodeInfo.model_validate(response.json())

    def set_config(self, new_config: dict[str, Any]) -> NodeSetConfigResponse:
        """Update configuration values of the node."""
        response = requests.post(
            f"{self.url}/config",
            json=new_config,
            timeout=60,
        )
        response.raise_for_status()
        return NodeSetConfigResponse.model_validate(response.json())

    def send_admin_command(self, admin_command: AdminCommands) -> bool:
        """Perform an administrative command on the node."""
        response = requests.post(f"{self.url}/admin/{admin_command}", timeout=10)
        response.raise_for_status()
        return AdminCommandResponse.model_validate(response.json())

    def get_resources(self) -> dict[str, ResourceDefinition]:
        """Get the resources of the node."""
        raise NotImplementedError(
            "get_resources is not implemented by this client",
        )
        # TODO: Implement get_resources endpoint

    def get_log(self) -> dict[str, Event]:
        """Get the log from the node"""
        response = requests.get(f"{self.url}/log", timeout=10)
        response.raise_for_status()
        return response.json()


def action_response_from_headers(headers: dict[str, Any]) -> ActionResult:
    """Creates an ActionResult from the headers of a file response"""

    return ActionResult(
        action_id=headers["x-madsci-action-id"],
        status=ActionStatus(headers["x-madsci-status"]),
        errors=json.loads(headers["x-madsci-errors"]),
        files=json.loads(headers["x-madsci-files"]),
        datapoints=json.loads(headers["x-madsci-datapoints"]),
    )


def process_file_response(rest_response: requests.Response) -> ActionResult:
    """Process a file rest response, saving files and getting headers"""
    response = action_response_from_headers(rest_response.headers)
    if response.files and len(response.files) == 1:
        file_key = next(iter(response.files.keys()))
        filename = response.files[file_key]
        with tempfile.NamedTemporaryFile(delete=False) as temp_file:
            temp_file.write(rest_response.content)
            temp_path = Path(temp_file.name)
        response.files[file_key] = temp_path
    elif response.files and len(response.files) > 1:
        with tempfile.NamedTemporaryFile(delete=False, suffix=".zip") as temp_zip:
            temp_zip.write(rest_response.content)
            temp_zip_path = Path(temp_zip.name)
        with ZipFile(temp_zip_path) as zip_file:
            for file_key in list(response.files.keys()):
                filename = response.files[file_key]
                with tempfile.NamedTemporaryFile(delete=False) as temp_file:
                    temp_file.write(zip_file.read(filename))
                    temp_path = Path(temp_file.name)
                response.files[file_key] = temp_path
    return response



"""Automated Tests for the MADSci cli's lab commands."""

import os
from pathlib import Path

import pytest
from click.testing import CliRunner
from madsci.client.cli import root_cli
from madsci.common.types.lab_types import LabDefinition


@pytest.fixture
def runner() -> CliRunner:
    """Fixture for creating a CliRunner instance."""
    return CliRunner()


@pytest.fixture
def temp_lab_file(tmp_path: Path) -> Path:
    """Fixture for creating a temporary lab definition file."""
    lab_file = tmp_path / "test_lab.lab.yaml"
    lab_definition = LabDefinition(
        name="test_lab",
        description="A test lab",
    )
    lab_definition.to_yaml(lab_file)
    return lab_file


@pytest.mark.skip(reason="Skipping lab CLI tests temporarily")
def test_lab_create(runner: CliRunner, tmp_path: Path) -> None:
    """Test creating a new lab definition."""
    result = runner.invoke(
        root_cli,
        [
            "lab",
            "create",
            "--name",
            "test_lab",
            "--description",
            "A test lab",
            "--path",
            str(tmp_path / "test_lab.lab.yaml"),
        ],
    )
    assert result.exit_code == 0
    assert (tmp_path / "test_lab.lab.yaml").exists()


@pytest.mark.skip(reason="Skipping lab CLI tests temporarily")
def test_lab_list(runner: CliRunner, temp_lab_file: Path) -> None:
    """Test listing all lab definitions."""
    os.chdir(temp_lab_file.parent)
    result = runner.invoke(root_cli, ["lab", "list"])
    assert result.exit_code == 0
    assert "test_lab" in result.output


@pytest.mark.skip(reason="Skipping lab CLI tests temporarily")
def test_lab_info(runner: CliRunner, temp_lab_file: Path) -> None:
    """Test getting information about a specific lab definition."""
    os.chdir(temp_lab_file.parent)
    result = runner.invoke(root_cli, ["lab", "--path", str(temp_lab_file), "info"])
    assert result.exit_code == 0
    assert "test_lab" in result.output


@pytest.mark.skip(reason="Skipping lab CLI tests temporarily")
def test_lab_delete(runner: CliRunner, temp_lab_file: Path) -> None:
    """Test deleting a lab definition."""
    os.chdir(temp_lab_file.parent)
    result = runner.invoke(
        root_cli, ["lab", "--path", str(temp_lab_file), "delete", "--yes"]
    )
    assert result.exit_code == 0
    assert not temp_lab_file.exists()


@pytest.mark.skip(reason="Skipping lab CLI tests temporarily")
def test_lab_validate(runner: CliRunner, temp_lab_file: Path) -> None:
    """Test validating a lab definition file."""
    os.chdir(temp_lab_file.parent)
    result = runner.invoke(root_cli, ["lab", "--path", str(temp_lab_file), "validate"])
    assert result.exit_code == 0
    assert "test_lab" in result.output


@pytest.mark.skip(reason="Skipping lab CLI tests temporarily")
def test_lab_commands(runner: CliRunner, temp_lab_file: Path) -> None:
    """Test adding a command to a lab definition."""
    os.chdir(temp_lab_file.parent)
    # * Test creating a new command
    result = runner.invoke(
        root_cli,
        [
            "lab",
            "--path",
            str(temp_lab_file),
            "add-command",
            "--name",
            "test_command",
            "--command",
            'echo "Hello, World!"',
        ],
    )
    assert result.exit_code == 0
    # * Test deleting the command
    result = runner.invoke(
        root_cli,
        [
            "lab",
            "--path",
            str(temp_lab_file),
            "delete-command",
            "test_command",
            "--yes",
        ],
    )
    assert result.exit_code == 0
    assert "Deleted command" in result.output
